/*
 * Micro-C ROBOTS - Robot header file
 *
 * Copyright 1999-2005 Dave dunfield
 * All rights reserved.
 */

/* These constants should match the ones in MCROBOTS.C */
#define	X_ARENA		128			/* Width of arena (combat area) */
#define	Y_ARENA		64			/* Height of arena (combat area) */
#define	MAX_DIR		16			/* Number of directions available */
#define	MAX_SPEED	10			/* Robot maximum speed */

/*
 * Control robot motion, 'direction' is 0 to halt robot, or
 * 1-MAX_DIR to select one of MAX_DIR compass points. If already
 * in motion, robot will decelerate until it stops, then accelerate
 * in the new direction until it reaches MAX_SPEED.
 */
void set_direction(direction) asm
{
	LD	2,S
	OUT	1
}

/*
 * Fire a shot in indicated direction: 0-(MAX_DIR-1) to select
 * one of 16 compass points. Shots will travel independantly of
 * the robot, and will cause damage to any robots within range
 * of the bullet. Bullets travel at a constant speed, however the
 * damage caused decreases with the distance travelled.
 */
void shoot(direction) asm
{
	LD  2,S
	OUT 2
}

/*
 * Warp transport! - Instantly move your robot to any location in the arena.
 * NOTE!!! Uses so much energy, it disables your gun for FIVE TIMES the
 * normal inter-shot delay.
 */
void warp(x, y) asm
{
	LD	4,S			; Get X position
	SHL	#8			; Into high byte
	OR	2,S			; Include Y in low byte
	OUT	3			; Perform warp function
}

/*
 * Add a beep to the sound queue
 */
void beep(f) asm
{
	LD	2,S
	OUT	4
}

/*
 * Get current robot ID (1-n)
 */
int get_id() asm
{
	IN	0
}

/*
 * Get our current direction of travel, returns value the same
 * as passed to "set_direction()".
 */
int get_direction() asm
{
	IN	1
}

/*
 * Get current robot speed: 0-MAX_SPEED
 */
int get_speed() asm
{
	IN	2
}

/*
 * Get our current 'X' position: 0-(X_ARENA-1)
 */
int get_x() asm
{
	IN 3
}

/*
 * Get our current 'Y' position: 0-(Y_ARENA-1)
 */
int get_y() asm
{
	IN	4
}

/*
 * Get our current damage level. This begins at 0, and will
 * increase every time we get hit, or collide with a wall or
 * another robot. When it reaches a preset limit, the robot
 * dies.
 */
int get_damage() asm
{
	IN	5
}

/*
 * Get current shot delay counter. This indicates how many ticks
 * remain before you can take another shot. A return value of 0
 * indicates that your gun is ready, and you can shoot now.
 */
int get_shot() asm
{
	IN	6
}

/*
 * Scan for a robot at a given direction. Value is: 0-(MAX_DIR-1),
 * and represents the 16 compass points. If a robot is found within
 * scanning range, it's id (1-#robots) is returned, otherwise a 0
 * is returned.
 */
int scan(direction) asm
{
	LD	2,S
	IN	7
}
