*
* CFLEA VM interpreter for 8051/52 host processors using standard memory.
*
* Copyright 1994-2005 Dave Dunfield
* All rights reserved.
*
* 805x register usage:
*   R0   = User Stack Pointer	(SP)
*   R1:2 = Operand/Address holding register
*   R3:4 = User Accumulator	(ACC)
*   R5:6 = User Index register	(I)
*   R7   = Misc temporary storage
*   DPTR = User Program Counter	(PC)
*
* Assemble with Dunfield ASM51: asm51 VM51
*
?R0	EQU	0		Stack pointer
?R1	EQU	?R0+1		Operand register (LOW)
?R2	EQU	?R0+2		Operand register (HIGH)
?R3	EQU	?R0+3		Accumulator (LOW)
?R4	EQU	?R0+4		Accumulator (HIGH)
?R5	EQU	?R0+5		Index register (LOW)
?R6	EQU	?R0+6		Index register (HIGH)
?R7	EQU	?R0+7		Misc regsiter
*
	ORG	$8000
	MOV	SP,#7		Set initial System SP
	MOV	R0,#$80		Set initial User SP
	MOV	DPTR,#$0100	Set initial User PC
*
******************************
* Fetch next opcode & decode *
******************************
*
fetchm	ACALL	readmem		Read opcode from PC
fetch1	MOV	B,A		Save opcode
*
* Standard memory reference instructions
*
	CJNE	A,#$98,*+3	>= 98?
	JNC	fetch5		Yes, try next group
	ACALL	getadr		Read addressing mode
	MOV	A,B		Get value back
	ANL	A,#%00000111	Immediate access?
	JNZ	fetch3		No, no special case
	ACALL	readnxt		Get LOW operand
	MOV	R1,A		Save value
	MOV	R2,#0		Assume BYTE
	MOV	A,B		Get operand back
	JB	A.3,fetch2	BYTE access only
	ACALL	readnxt		Get HIGH byte
	MOV	R2,A		Save value
fetch2	PUSH	DPL		Save LOW PC
	PUSH	DPH		Save HIGH PC
	AJMP	fetch4		And continue
fetch3	PUSH	DPL		Save LOW PC
	PUSH	DPH		Save HIGH PC
	MOV	DPL,R1		Set LOW address
	MOV	DPH,R2		Set HIGH address
	ACALL	readmem		Read LOW value
	MOV	R1,A		Set LOW operand
	MOV	R2,#0		Assume BYTE access
	MOV	A,B		Get opcode back
	JB	A.3,fetch4	BYTE access
	ACALL	readnxt		Read HIGH value
	MOV	R2,A		Set HIGH operand
fetch4	MOV	A,B		Get opcode back
	RR	A		Shift opcode
	RR	A		Over into
	RR	A		Correct position
	ANL	A,#%00011110	Mask unused bits
	MOV	DPTR,#OPTAB1	Opcode table 1
	JMP	[A+DPTR]	Execute handler
*
* Special memory reference instructions
*
fetch5	CJNE	A,#$C8,*+3	Is it >= C0?
	JNC	fetch6		Yes, try next group
	ACALL	getadr		Get operand address
	PUSH	DPL		Save DPL
	PUSH	DPH		Save DPH
	MOV	A,B		Get opcode back
	ADD	A,#-$98		Adjust to binary
	RR	A		Shift opcode over
	RR	A		Into proper place
	ANL	A,#%00001110	Mask unused bits
	MOV	DPTR,#OPTAB2	Opcode table 2
	JMP	[A+DPTR]	Execute handler
*
* Other opcodes
*
fetch6	PUSH	DPL		Save LOW PC
	PUSH	DPH		Save HIGH PC
	ADD	A,#-$C8		Adjust for offset
	CLR	C		Zero carry
	RLC	A,1		X2 for word entries
	MOV	DPTR,#OPTAB3	Opcode table 3
	JMP	[A+DPTR]	Execute handler
*
* Get address into operand register
*
getadr	ANL	A,#%00000111	Save only addressing mode
	MOV	R7,A		Save for later test
* Immediate addressing mode
	JNZ	geta1		Not immediate
	MOV	R1,DPL		Save LOW byte
	MOV	R2,DPH		Save HIGH byte
	RET
* Extended addressing mode
geta1	DJNZ	R7,geta2	Not extended
	ACALL	readnxt		Get opcode from PC
	MOV	R1,A		Save LOW byte
	ACALL	readnxt		Get opcide from PC
	MOV	R2,A		Save HIGH byte
	RET
* Indirect through index register (No offset)
geta2	DJNZ	R7,geta3	Not indexed
	MOV	?R1,R5		Get Index LOW
	MOV	?R2,R6		Get Index HIGH
	RET
* Indirect through index register (with offset)
geta3	DJNZ	R7,geta4	Not indexed
	ACALL	readnxt		Get offset from PC
	ADD	A,R5		Add to index
	MOV	R1,A		Set LOW operand
	MOV	A,R6		Get high value
	ADDC	A,#0		Adjust for overflow
	MOV	R2,A		Set HIGH operand
	RET
* Indirect through stack pointer (with offset)
geta4	DJNZ	R7,geta5	Not stack relative
	ACALL	readnxt		Get offset from PC
	ADD	A,R0		Adjust for stack
	MOV	R1,A		Set LOW operand
geta41	MOV	R2,#0		Clear HIGH operand
	RET
* Top of stack (remove)
geta5	DJNZ	R7,geta6	Not Top of stack
	MOV	?R1,R0		Point to TOS
	INC	R0		Advance SP
	INC	R0		Advance SP
	SJMP	geta41		Zero HIgh & exit
* Indirect through top of stack (remove)
geta6	DJNZ	R7,geta7	Not Indirect on top
	MOV	?R1,[R0]	Get LOW operand
	INC	R0		Advance stack
	MOV	?R2,[R0]	Get HIGH operand
	INC	R0		Advance stack
	RET
* Indirect through top of stack (leave on)
geta7	MOV	?R1,[R0]	Set LOW operand
	INC	R0		Advance stack
	MOV	?R2,[R0]	Set HIGH operand
	DEC	R0		Reset stack
	RET
*
* General Memory Addressing Instruction Handlers
*
OPTAB1	AJMP	?LD
	AJMP	?ADD
	AJMP	?SUB
	AJMP	?MUL
	AJMP	?DIV
	AJMP	?AND
	AJMP	?OR
	AJMP	?XOR
	AJMP	?CMP
	AJMP	?LDI
?LD	MOV	R3,?R1		Get LOW value
	MOV	A,R2		Get HIGH value
	AJMP	?exit1		Save & Exit
?INC	MOV	R1,#1		Get 1
	MOV	R2,#0		Zero high
?ADD	MOV	A,R3		Get LOW ACC
	ADD	A,R1		Add LOW operand
	MOV	R3,A		Resave LOW ACC
	MOV	A,R4		Get HIGH ACC
	ADDC	A,R2		Add HIGH operand
?exit1	MOV	R4,A		Resave HICH ACC
?exit2	POP	DPH		Restore PC
	POP	DPL		Restore PC
fetchn	ACALL	readnxt		Read next byte
	AJMP	fetch1		Fetch next opcide
?DEC	MOV	R1,#1		Get 1
	MOV	R2,#0		Zero high
?SUB	CLR	C		Zero carry
	MOV	A,R3		Get LOW ACC
	SUBB	A,R1		Subtract LOW operand
	MOV	R3,A		Resave
	MOV	A,R4		Get HIGH ACC
	SUBB	A,R2		Subtract HIGH operand
	AJMP	?exit1		Resave & exit
?MUL	MOV	B,R4		Get HI1
	MOV	A,R1		Get LO2
	MUL			Multiply HI1 * LO2
	MOV	R7,A		Save partial product
	MOV	B,R3		Get LO1
	MOV	A,R2		Get HI2
	MUL			Multiply LO1 * HI2
	MOV	R2,A		Save partial product
	MOV	B,R3		Get LO1
	MOV	A,R1		Get LO2
	MUL			Multiply LO1 * LO2
	XCH	A,B		Save low & get high
	ADD	A,R2		Add in carry from LO1 * HI2
	ADD	A,R7		Add in carry from HI1 * LO2
	MOV	R3,B		Save LOW result
	AJMP	?exit1		And exit
?DIV	MOV	DPL,R5		Save index HIGH
	MOV	DPH,?R6		Save index LOW
	MOV 	R6,#0		Begin with zero
	MOV 	R5,#0		"" ""
	MOV	R7,#17		16+1 bit shift
?div1	CLR	C		No carry in
?div2	MOV	A,R3		Shift parm1 left, C = carry out
	RLC	A		Do the shift
	MOV	R3,A		Resave
	MOV	A,R4		Get high half
	RLC	A		Do the shift
	MOV	R4,A		Resave
	DJNZ	R7,?div3	Not done, keep going
* Get result and exit
	MOV	R1,?R5		Save remainder LOW
	MOV	R2,?R6		Save remainder HIGH
	MOV	R5,DPL		Restore index LOW
	MOV	R6,DPH		Restore index HIGH
	AJMP	?exit1		And go home
* Compare registers & shift in appropriate bit
?div3	MOV	A,R5		Shift carry into secondary result
	RLC	A		Do the shift
	MOV	R5,A		Resave
	MOV	A,R6		Get high half
	RLC	A		Do the shift
	MOV	R6,A		Resave
	CLR	C		Clear carry for test
	MOV	A,R5		Get secondary LOW
	SUBB	A,R1		Test against primary LOW
	MOV	B,A		Save for later store
	MOV	A,R6		Get secondary HIGH
	SUBB	A,R2		Test against poimary high
	JC	?div1		Still smaller, shift in 0
	MOV	R5,B		Set new secondary high
	MOV	R6,A		Set new secondary low
	SETB	C		Set '1' in register
	SJMP	?div2		Shift in one
?AND	MOV	A,R3		Get LOW ACC
	ANL	A,R1		AND with LOW operand
	MOV	R3,A		Resave
	MOV	A,R4		Get HIGH ACC
	ANL	A,R2		AND with HIGH operand
	AJMP	?exit1		Resave & exit
?OR	MOV	A,R3		Get LOW ACC
	ORL	A,R1		OR with LOW operand
	MOV	R3,A		Resave
	MOV	A,R4		Get HIGH ACC
	ORL	A,R2		OR with HIGH operand
	AJMP	?exit1		Resave & exit
?XOR	MOV	A,R3		Get LOW ACC
	XRL	A,R1		XOR with LOW operand
	MOV	R3,A		Resave
	MOV	A,R4		Get HIGH ACC
	XRL	A,R2		XOR with HIGH operand
	AJMP	?exit1		Resave & exit
?CMP	CLR	C		Zero carry
	MOV	A,R3		Get LOW ACC
	SUBB	A,R1		Subtract LOW result
	MOV	R3,A		Resave
	MOV	A,R4		Get HIGH ACC
	SUBB	A,R2		Subtract HIGH operand
	MOV	R4,A		Resave
	MOV	A.6,C		A.7 = Sign, A.6=Carry
	MOV	C,PSW.2		Get Overflow
	MOV	A.5,C		A.5 = Overflow
	MOV	A.4,C		A.4 = S ^ O
	SETB	A.3		A.3 = Zero (Assume OK)
	JNB	A.7,cmp1	No XOR
	CPL	A.4		Toggle bit
cmp1	MOV	R7,A		Save flags for later
	MOV	A,R3		Get ACC LOW
	ORL	A,R4		Test for ZERO (EQUAL)
	JZ	ret1		Yes, return 1
	MOV	A,R7		Get value back
	CLR	A.3		Clear 'Z' (not zero)
	MOV	R7,A		Resave
ret0	CLR	A		Zero ACC
	MOV	R3,A		Zero ACC LOW
	AJMP	?exit1
ret1	CLR	A		Zero Acc
	MOV	R3,#1		Set LOW to 1
	AJMP	?exit1		Save HIGH & exit
* A.7=Sign, A.6=Carry, A.5=Ovfl, A.4=S^O, A.3=Zero
?LE	MOV	A,R7
	JB	A.3,ret1
?LT	MOV	A,R7
	JB	A.4,ret1
	SJMP	ret0
?GT	MOV	A,R7
	JB	A.3,ret0
?GE	MOV	A,R7
	JNB	A.4,ret1
	SJMP	ret0
?ULE	MOV	A,R7
	JB	A.3,ret1
?ULT	MOV	A,R7
	JB	A.6,ret1
	SJMP	ret0
?UGT	MOV	A,R7
	JB	A.3,ret0
?UGE	MOV	A,R7
	JNB	A.6,ret1
	SJMP	ret0
?NOT	MOV	A,R3		Get LOW
	ORL	A,R4		Zero?
	JZ	ret1		Yes, send back one
	SJMP	ret0		No, return 0
?LDI	MOV	R5,?R1		Copy LOW value
	MOV	R6,?R2		Copy HIGH value
	AJMP	?exit2		And proceed
*
* Special memory addressing instructions
*
OPTAB2	AJMP	?LDI		(Actually LEAI - Handler is same)
	AJMP	?ST
	AJMP	?STB
	AJMP	?STI
	AJMP	?SHR
	AJMP	?SHL
?ST	MOV	DPL,R1		Get LOW memory address
	MOV	DPH,R2		Get HIGH memory address
	MOV	A,R3		Get LOW accumulator
	ACALL	wrimem		Write to memory
	MOV	A,R4		Get HIGH value
	ACALL	wrinxt		Write to memory
	AJMP	?exit2		And exit
?STB	MOV	DPL,R1		Get LOW memory address
	MOV	DPH,R2		Get HIGH memory address
	MOV	A,R3		Get LOW accumulator
	ACALL	wrimem		Write to memory
	AJMP	?exit2
?STI	MOV	DPL,R1		Get LOW memory address
	MOV	DPH,R2		Get HIGH memory address
	MOV	A,R5		Get LOW accumulator
	ACALL	wrimem		Write to memory
	MOV	A,R6		Get HIGH value
	ACALL	wrinxt		Write to memory
	AJMP	?exit2		And exit
?SHR	MOV	A,B		Get addressing mode
	ANL	A,#%00000111	Immediate mode?
	JNZ	?shr1		No, fetch from external
	POP	DPH		Restore PC high
	POP	DPL		Restore PC low
	ACALL	readnxt		Read next value
	AJMP	?shr2		And proceed
?shr1	MOV	DPL,R1		Get LOW address
	MOV	DPH,R2		Get HIGH address
	ACALL	readmem		Read memory value
	POP	DPH		Restore HIGH
	POP	DPL		Restore LOW
?shr2	MOV	R7,A		Set counter
	JZ	?shr4		Zero, exit now
?shr3	CLR	C		Zero carry in
	MOV	A,R4		Get HIGH
	RRC	A		Rotate right
	MOV	R4,A		Resave
	MOV	A,R3		Get LOW
	RRC	A		Rotate right
	MOV	R3,A		Resave
	DJNZ	R7,?shr3	Do them all
?shr4	AJMP	fetchn		And fetch next
?SHL	MOV	A,B		Get addressing mode
	ANL	A,#%00000111	Immediate mode?
	JNZ	?shl1		No, fetch from external
	POP	DPH		Restore PC high
	POP	DPL		Restore PC low
	ACALL	readnxt		Read next value
	AJMP	?shl2		And proceed
?shl1	MOV	DPL,R1		Get LOW address
	MOV	DPH,R2		Get HIGH address
	ACALL	readmem		Read memory value
	POP	DPH		Restore HIGH
	POP	DPL		Restore LOW
?shl2	MOV	R7,A		Set counter
	JZ	?shl4		Zero, exit now
?shl3	CLR	C		Zero carry in
	MOV	A,R3		Get HIGH
	RLC	A		Rotate left
	MOV	R3,A		Resave
	MOV	A,R4		Get LOW
	RLC	A		Rotate left
	MOV	R4,A		Resave
	DJNZ	R7,?shl3	Do them all
?shl4	AJMP	fetchn		And fetch next
*
* Other opcodes
*
OPTAB3	AJMP	?LT		C8
	AJMP	?LE		C9
	AJMP	?GT		CA
	AJMP	?GE		CB
	AJMP	?ULT		CC
	AJMP	?ULE		CD
	AJMP	?UGT		CE
	AJMP	?UGE		CF
	AJMP	?JMP		D0
	AJMP	?JZ		D1
	AJMP	?JNZ		D2
	AJMP	?SJMP		D3
	AJMP	?SJZ		D4
	AJMP	?SJNZ		D5
	AJMP	?IJMP		D6
	AJMP	?SWITCH		D7
	AJMP	?CALL		D8
	AJMP	?RTS		D9
	AJMP	?ALLOC		DA
	AJMP	?FREE		DB
	AJMP	?PUSHA		DC
	AJMP	?PUSHI		DD
	AJMP	?TAS		DE
	AJMP	?TSA		DF
	AJMP	?CLR		E0
	AJMP	?COM		E1
	AJMP	?NEG		E2
	AJMP	?NOT		E3
	AJMP	?INC		E4
	AJMP	?DEC		E5
	AJMP	?TAI		E6
	AJMP	?TIA		E7
	AJMP	?ADAI		E8
	AJMP	?ALT		E9
	AJMP	?OUT		EA
	AJMP	?IN		EB
	AJMP	?NATIVE		EC
	AJMP	?exit2		ED
	AJMP	?exit2		EE
	AJMP	?exit2		EF
	AJMP	?exit2		F0
	AJMP	?exit2		F1
	AJMP	?exit2		F2
	AJMP	?exit2		F3
	AJMP	?exit2		F4
	AJMP	?exit2		F5
	AJMP	?exit2		F6
	AJMP	?exit2		F7
	AJMP	?exit2		F8
	AJMP	?exit2		F9
	AJMP	?exit2		FA
	AJMP	?exit2		FB
	AJMP	?exit2		FC
	AJMP	?exit2		FD
	AJMP	?exit2		FE
	AJMP	TRAP		FF
?JMP	POP	DPH		Restore HIGH
	POP	DPL		Restore LOW
	ACALL	readnxt		Get LOW address
	MOV	R7,A		Save it
	ACALL	readnxt		Get HIGH address
?jmp1	MOV	DPH,A		Save HIGH
	MOV	DPL,R7		Save LOW
	AJMP	fetchm		And proceed
?JZ	MOV	A,R3		Get ACC low
	ORL	A,R4		Zero?
	JZ	?JMP
?skip2	POP	DPH		Restore HIGH
	POP	DPL		Restore LOW
	ACALL	readnxt		Get LOW
	ACALL	readnxt		Get HIGH
	AJMP	fetchn		And proceed
?JNZ	MOV	A,R3		Get LOW
	ORL	A,R4		Zero ?
	JNZ	?JMP		No, do jump
	SJMP	?skip2		Yes, don't jump
?SJMP	POP	DPH		Restore HIGH
	POP	DPL		Restore LOW
	MOV	R2,#0		Assume positive
	ACALL	readnxt		Get LOW address
	MOV	R1,A		Save LOW
	JNB	A.7,?sjmp1	Assumption correct
	DEC	R2		Convert to negative
?sjmp1	MOV	A,DPL		Get LOW
	ADD	A,R1		Add LOW
	MOV	DPL,A		Resave
	MOV	A,DPH		Get HIGH
	ADDC	A,R2		Adjust
	MOV	DPH,A		Resave
	AJMP	fetchm		And proceed
?SJZ	MOV	A,R3		Get ACC
	ORL	A,R4		Zero?
	JZ	?SJMP		Do the jump
?skip1	POP	DPH		Restore HIGH
	POP	DPL		Restore LOW
	ACALL	readnxt		Skip 1 byte
	AJMP	fetchn
?SJNZ	MOV	A,R3		Get ACC
	ORL	A,R4		Zero?
	JNZ	?SJMP		No, Do jump
	SJMP	?skip1		Proceed
?IJMP	POP	DPH		Restore HIGH
	POP	DPL		Restore LOW
	MOV	DPL,R3		Set LOW
	MOV	DPH,R4		Set HIGH
	AJMP	fetchm		And proceed
?SWITCH	POP	DPH
	POP	DPL
	MOV	DPL,R5		Get INDEX LOW
	MOV	DPH,R6		Get INDEX HIGH
	ACALL	readmem		Read a byte
?swit1	MOV	R1,A		Save for later
	ACALL	readnxt		Read next value
	MOV	R7,A		Save for compare
	ORL	A,R1		End of table
	JZ	?swit4		Yes, do default
	ACALL	readnxt		Get LOW value
	MOV	B,A		Save
	ACALL	readnxt		Get HIGH value
	CJNE	A,?R4,?swit3	Not same, skip
	MOV	A,B		Get LOW
	CJNE	A,?R3,?swit3	Not same, skip
?swit2	MOV	DPL,R1		Set LOW address
	MOV	DPH,R7		Set HIGH address
	AJMP	fetchm		And proceed
?swit3	ACALL	readnxt		Get next byte
	SJMP	?swit1		And proceed
?swit4	ACALL	readnxt		Get LOW address
	MOV	R1,A		Save
	ACALL	readnxt		Get HIGH address
	MOV	R7,A		Save
	SJMP	?swit2		Do the jump
?CALL	POP	DPH
	POP	DPL
	ACALL	readnxt		Get LOW address
	MOV	R7,A		Save LOW address
	ACALL	readnxt		Get HIGH address
	DEC	R0		Backup stack
	MOV	[R0],DPH	Save HIGH
	DEC	R0		Backup stack
	MOV	[R0],DPL	Save LOW
	AJMP	?jmp1		Do the jump
?RTS	POP	DPH
	POP	DPL
	MOV	DPL,[R0]	Get LOW
	INC	R0		Advance stack
	MOV	DPH,[R0]	Get HIGH
	INC	R0		Advance stack
	AJMP	fetchm		And proceed
?ALLOC	POP	DPH
	POP	DPL
	ACALL	readnxt		Get offset
	XCH	A,R0		Swap out
	CLR	C		Zero carry
	SUBB	A,R0		Compute new value
?all1	MOV	R0,A		And proceed
	AJMP	fetchn
?FREE	POP	DPH
	POP	DPL
	ACALL	readnxt		Get offset
	ADD	A,R0		Release space
	SJMP	?all1		And proceed
?PUSHA	DEC	R0		Backup stack
	MOV	[R0],?R4	Write HIGH
	DEC	R0		Backup stack
	MOV	[R0],?R3	Write LOW
	AJMP	?exit2
?PUSHI	DEC	R0		Backup stack
	MOV	[R0],?R6	Write HIGH
	DEC	R0		Backup stack
	MOV	[R0],?R5	Write LOW
	AJMP	?exit2
?TAS	MOV	?R0,R3		Copy value
	AJMP	?exit2
?TSA	MOV	R3,?R0		Get LOW
	CLR	A		Zero HIGH
	AJMP	?exit1
?CLR	CLR	A		Get ZERO
	MOV	R3,A		Zero LOW
	AJMP	?exit1
?NEG	CLR	A		Get zero
	CLR	C		Zero carry
	SUBB	A,R3		Subtract LOW
	MOV	R3,A		Resave
	CLR	A		Get zero
	SUBB	A,R4		Subtract high
	AJMP	?exit1
?COM	MOV	A,R3		Get LOW
	CPL	A		Complement
	MOV	R3,A		Resave
	MOV	A,R4		Get HIGH
	CPL	A		Complement
	AJMP	?exit1
?TAI	MOV	R5,?R3		Copy LOW
	MOV	R6,?R4		Copy HIGH
	AJMP	?exit2
?TIA	MOV	R3,?R5		Copy LOW
	MOV	R4,?R6		Copy HIGH
	AJMP	?exit2
?ADAI	MOV	A,R5		Get LOW index
	ADD	A,R3		Add LOW ACC
	MOV	R5,A		Resave
	MOV	A,R6		Get HIGH index
	ADDC	A,R4		Add HIGH ACC
	MOV	R6,A		Resave
	AJMP	?exit2		And proceed
?ALT	MOV	R3,?R1		Copy LOW
	MOV	R4,?R2		Copy HIGH
	AJMP	?exit2
*
* Write to output port
*
?OUT	POP	DPH
	POP	DPL
	ACALL	readnxt		Get port ID
	JNZ	out1		Not SERIAL
	JNB	SCON.1,*	Wait for TX ready
	CLR	SCON.1		Reset bit
	MOV	SBUF,R3		Write value
	AJMP	fetchn
out1	MOV	R7,A		Set counter
	DJNZ	R7,out2		Not PORT1
	MOV	P1,R3		Write P1 value
	AJMP	fetchn
out2	DJNZ	R7,in0		Not PORT2
	MOV	P3,R3		Write P3 value
	AJMP	fetchn
*
* Read from port
*
?IN	CLR	A		Get zero
	MOV	R4,A		Zero HIGH
	MOV	R3,A		Zero LOW
	POP	DPH
	POP	DPL
	ACALL	readnxt		Get port value
	JNZ	in1		Not SERIAL
	JNB	SCON.0,in0	No data ready
	CLR	SCON.0		Reset bit
	MOV	A,SBUF		Get data
	MOV	R3,A		Save value
in0	AJMP	fetchn
in1	MOV	R7,A		Save counter
	DJNZ	R7,in3		Not PORT1
	MOV	R3,P1		Read P1 value
	AJMP	fetchn
in3	DJNZ	R7,in0		Not PORT2
	MOV	R3,P3		Read P3 value
	AJMP	fetchn
*
* Execute native code (Direct memory only)
*
?NATIVE	MOV	A,R3		Copy ACC low
	MOV	B,R4		Copy ACC high
	MOV	DPL,R5		Copy INDEX low
	MOV	DPH,R6		Copy INDEX high
	RET			Execute user code
*
* Halt simulation in infinite loop
*
TRAP	POP	DPH
	POP	DPL
	SJMP	*
	AJMP	fetchn
*
* Read a location from memory (DPTR)
*
readmem
* Read next location from memory (DPTR)
readnxt	MOV	A,DPH		Get HIGH value
	JNZ	readn1		Use EXTERNAL memory
	MOV	R2,?R0		Save SP
	MOV	R0,DPL		Get LOW address
	MOV	A,[R0]		Read value
	MOV	R0,?R2		Restore SP
	INC	DPTR		Skip to next
	RET
readn1	MOVX	A,[DPTR]	Read value
	INC	DPTR		Skip to next
	RET
*
* Write a location in memory (DPTR)
*
wrimem
* Write next location in memory (DPTR)
wrinxt	XCH	A,DPH		Get HIGH
	JNZ	wrin1		Use EXTERNAL memory
	XCH	A,DPH		Swap back
	MOV	R2,?R0		Save SP
	MOV	R0,DPL		Get LOW
	MOV	[R0],A		Write to memory
	MOV	R0,?R2		Restore SP
	INC	DPTR		Skip to next
	RET
wrin1	XCH	A,DPH		Swap back
	MOVX	[DPTR],A	Write it
	INC	DPTR		Skip to next
	RET
