*
* 6800/01/02/08/11/12 CFLEA VM interpreter with integrated debugger.
*
* Copyright 1994-2005 Dave Dunfield
* All rights reserved.
*
* CFLEA debugger commands:
* ------------------------
* B [addr]	- Set breakpoint
* C R value	- Change register value
* D addr,[addr]	- Dump memory
* E addr	- Edit memory
* G [addr]	- Go (execute)
* K addr	- Kill breakpoint
* R		- Display regsiters
* S		- Single step one instruction
* U addr,[addr]	- Unassemble memory
* X		- Exit to system monitor
* Z		- Reset CFLEA
*
* Assemble with Dunfield ASM00/01/11/12: asm00 VM68
*
* Note: When using ASM12, you will get an "out of range" error on one of
*       the symbols in the disassembler. You need to reduce the code size
*       of that block by 1 byte. This can be easily done by using 68HC12
*       enhanced instructions. eg: replace LDAA 0,X : INX with LDAA 1,X+
*
* Global system memory addresses
CODE	EQU	$1000		Place code here
STACK	EQU	$0EDE		CFLEA system stack
ISP	EQU	$9FFF		Initial user SP
IPC	EQU	$4003		Initial user PC (Skip CALL)
*
MONIT	EQU	$F37E		Monitor reentry point
OUTSER	EQU	$E75D		Write A to serial port
INSER	EQU	$E7D0		Read A from serial port
*
* CFLEA page 0 variables
*
	ORG	$0040		Skip monitor area
TEMP	RMB	2		Temp location
ALTVAL	RMB	2		Alt value (for DIV)
ACC	RMB	2		Virtual accumulator
IDX	RMB	2		Virtual index register
SP	RMB	2		Virtual stack pointer
PC	RMB	2		Virtual program counter
OPRADR	RMB	2		Memory/Operand address
BRKTAB	RMB	4*3		Breakpoint table
BRKEND	EQU	*		Marks end of table
* Page 0 variables used by library
Motion	RMB	1		Motion detector
*
* Main CFLEA entry points
*
	ORG	CODE		Code goes here
* CODE+$0000 - Cold interactive start (debugger)
DEBUG	BRA	BEGIN		Execute here
* CODE+$0002 - Cold RUN (execute program)
COLDR	BSR	INITCF		Initialize CFLEA
* CODE+$0004 - Continue execution (after NATIVE)
WARMR	STAA	ACC		Save ACCH
	STAB	ACC+1		Save ACCL
	STX	IDX		Save X
	PULA			Get high PC
	STAA	PC		Save high PC
	PULA			Get low PC
	STAA	PC+1		Save LOW PC
	LDS	#STACK		Insure stack is OK
	LDX	OPRADR		Get last execution
	CPX	#CMDRET		Were we stepping?
	BEQ	COMMAND		Yes, get next command
	JMP	RUN		Execute code
*
* Initialize the CFLEA virtual processor for interactive session
*
INITCF	LDX	#ALTVAL		First memory location
init1	CLR	0,X		Zero one byte
	INX			Advance
	CPX	#BRKEND		End of table?
	BNE	init1		Do entire table
	LDX	#ISP		Initial stack location
	STX	SP		Save it
	LDX	#IPC		Initial program counter
	STX	PC		Save it
	RTS
*
* Interactive entry point
*
BEGIN	BSR	INITCF		Initialize CFLEA
	LDX	#HELLO		Welcome message
	JSR	WRSTR		Output
*
* Main command handling loop
*
COMMAND	LDS	#STACK		Reset stack incase error
	JSR	WRLFCR		New line
	LDAA	#'>'		Prompt character
	JSR	WRCHR		Output
	JSR	RDCHRE		Get character
	PSHA			Save
	JSR	WRSPACE		Space over
	PULA			Restore
	LDX	#CMDTAB-3	Point to command table
Comm1	INX			Skip cmd char
	INX			Skip high address
	INX			Skip low address
	TST	0,X		End of table?
	BEQ	ERROR		Yes, report error
	CMPA	0,X		Is this it?
	BNE	Comm1		No, keep looking
	LDX	1,X		Get address
	JSR	0,X		Execute command
CMDRET	BRA	COMMAND		Next command
* Command handler table
CMDTAB	FCB	'D'		Dump memory
	FDB	DUMP
	FCB	'E'		Edit memory
	FDB	EDIT
	FCB	'R'		Register display
	FDB	REGDIS
	FCB	'C'		Change register
	FDB	CHGREG
	FCB	'U'		Unassemble
	FDB	UNASM
	FCB	'S'		Single-step
	FDB	STEP
	FCB	'G'		Go?
	FDB	GO
	FCB	'B'		Breakpoint
	FDB	BREAKPT
	FCB	'K'		Kill breakpoint
	FDB	KILLB
	FCB	'Z'		Reset CPU
	FDB	BEGIN
	FCB	'X'		Exit to monitor
	FDB	MONIT
	FCB	'?'		Help request
	FDB	HELP
	FCB	0		End of table
*
* An error has occured
*
ERROR	JSR	WRSPACE		Space over
	LDAA	#'?'		Error prompt
	JSR	WRCHR		Output
	BRA	COMMAND		Next command
*
* Dump memory in HEX/ASCII
*
DUMP	LDX	#$00FF		Default to 256 bytes
	JSR	RDADDR		Get address to dump
dump1	JSR	WRLFCR		New line
	JSR	WRADDR		Output address
	LDAB	#16		16 bytes/line
dump2	JSR	WRSPACE		Space over
	LDAA	0,X		Get byte from line
	INX			Advance to next
	JSR	WRHEX		Output
	DECB			Reduce count
	BNE	dump2		Do whole line
	JSR	WRSP3		Space over
	LDX	TEMP		Restore address
	LDAB	#16		16 bytes/line
dump3	LDAA	0,X		Get byte from line
	INX			Advance to next
	CMPA	#' '		Printable?
	BLO	dump4		No, convert
	CMPA	#'~'		Printable?
	BLS	dump5		Yes, output
dump4	LDAA	#'.'		Convert unprint to dot
dump5	JSR	WRCHR		Output character
	DECB			Reduce count
	BNE	dump3		Output full line
	BSR	CPXEND		End of range?
	BLO	dump1		No, go again
	RTS
*
* Compare X and OPRADR (C = <=)
*
CPXEND	DEX			Adjust
	STX	TEMP		Save temp
	INX			Fixup
	LDAA	TEMP+1		Get LOW
	SUBA	OPRADR+1	Subtract
	LDAA	TEMP		Get HIGH
	SBCA	OPRADR		Subtract high
cpxe1	RTS
*
* Edit memory
*
EDIT	JSR	RDWORD		Get address
	BCS	ERROR		Report error
	STAA	TEMP		Save high
	STAB	TEMP+1		Save LOW
edit1	JSR	WRLFCR		New line
	JSR	WRADDT		Output address
edit2	JSR	WRSPACE		Space over
	LDX	TEMP		Get address
	LDAA	0,X		Get byte
	JSR	WRHEX		Output
	LDAA	#'-'		Prompt
	JSR	WRCHR		Output
	JSR	RDHEX		Get hex value
	BCS	edit4		Error, try next
edit21	STAA	0,X		Write it
edit3	INX			Advance to next
	STX	TEMP		Save new
	LDAA	TEMP+1		Get address
	BITA	#%00000111	8 char limit
	BNE	edit2		Prompt on same line
	BRA	edit1		New line
edit4	CMPA	#$0D-'0'	End
	BEQ	cpxe1		Exit
	CMPA	#' '-'0'	Space?
	BNE	edit5		No, try next
	JSR	WRSPACE		Space over
	BRA	edit3		And do next
edit5	CMPA	#$08-'0'	Backspace?
	BNE	edit6		No, try next
	DEX			Backup
	STX	TEMP		Save new
	BRA	edit1		New line
edit6	CMPA	#$27-'0'	Quote?
	BNE	edit1		No, reprompt
	JSR	RDCHR		Get char
	BRA	edit21		Save & advance
*
* TRAP... return to monitor
*
_TRAP	LDX	PC		Get PC
	DEX			Backup to offender
	STX	PC		Resave
	CLRB			Indicate TRAP
* Remove installed breakpoints
	LDX	#BRKTAB		Get breakpoint table
trap2	STX	TEMP		Save for later
	LDAA	2,X		Get opcode
	LDX	0,X		Get address
	BEQ	trap3		No breakpoint
	STAA	0,X		Rewrite opcode
	CPX	PC		Is this it
	BNE	trap3		No, ignore
	DECB			Indicate BREAKPOINT
trap3	LDX	TEMP		Get address
	INX			Skip addressH
	INX			Skip addressL
	INX			Skip opcode
	CPX	#BRKEND		At end of table?
	BNE	trap2		Do them all
* Output message & register values
trap1	LDX	#TRAPMSG	Its a trap!
	TSTB			Breakpoint?
	BEQ	trap4		No, it really was a trap
	LDX	#BRKMSG		Its a breakpoint
trap4	JSR	WRSTR		Output
	BSR	REGDIS		Display registers
	JMP	COMMAND		And return
*
* Display registers
*
REGDIS	LDX	#RDISTXT	Point to string
regd1	JSR	WRSTR		Output string
	STX	TEMP		Save it
	LDX	0,X		Get address
	LDAA	0,X		Get LOW byte
	LDAB	1,X		Get HIGH byte
	LDX	TEMP		Restore
	JSR	WRWORD		Output
	INX			Skip HIGH
	INX			Skip LOW
	TST	0,X		End of list?
	BNE	regd1		Keep going
	RTS
*
* Output newline to console
*
WRLFCR	LDAA	#$0A		Get LF
	BSR	WRCHR		Output
	LDAA	#$0D		Get CD
	BRA	WRCHR
*
* Output a space to the console
*
WRSP3	BSR	WRSPACE		Output another space
WRSP2	BSR	WRSPACE		Output another space
WRSPACE	LDAA	#' '		Get space
	BRA	WRCHR		Output
*
* Write string(X) to console
*
WRSTR	LDAA	0,X		Get data byte
	ANDA	#%01111111	Mask high but
	BSR	WRCHR		Output
	DECB			Reduce count
	LDAA	0,X		Get data back
	INX			Skip to next
	BPL	WRSTR		Keep going
	RTS
*
* Write word(X) to console
*
WRADDR	STX	TEMP		Save for later
WRADDT	LDAA	TEMP		Get high
	LDAB	TEMP+1		Get LOW
*
* Write word(A:B) to console
*
WRWORD	BSR	WRHEX		Output A
	TBA			Get B
*
* Write byte in A to console in hex
*
WRHEX	PSHA			Save for later
	LSRA			Shift high
	LSRA			nibble over
	LSRA			into low
	LSRA			nibble
	BSR	WRNIB		Output nibble
	PULA			Restore byte
WRNIB	ANDA	#%00001111	Save ony low nibble
	ADDA	#'0'		Convert to ASCII
	CMPA	#'9'		In range?
	BLS	WRCHR		Yes, its OK
	ADDA	#7		Adjust for alphas
*
* Write character in 'A' to console
*
WRCHR	SEI			Disable interrupts
	JSR	OUTSER		Output character
	CLI			Re-enable interrupts
	RTS
*
* Read character, convert to UC
*
RDCHRE	BSR	RDCHR		Get char
	CMPA	#'a'		Lower case?
	BLO	rdche1		No, its ok
	ANDA	#$5F		Convert to UPPER
rdche1	RTS
*
* Read a character from console
*
RDCHR	SEI			Disable interrupts
	JSR	INSER		Read character
	CLI			Allow interrupt
	ANDA	#$7F
	RTS
*
* Read a nibble in HEX
*
RDNIB	BSR	RDCHR		Read a character
	SUBA	#'0'		Convert to ASCII
	BLO	rdnerr		Report error
	CMPA	#9		Numeric?
	BLS	rdn1		Yes, its OK
	ANDA	#$5F		Convert to upper
	SUBA	#7		Convert
	CMPA	#10		In range?
	BLO	rdnerr		No, report error
	CMPA	#15		In range?
	BHI	rdnerr		No, report error
rdn1	CLC			Indicate OK
	RTS
rdnerr	SEC			Indicate error
	RTS
*
* Read a byte in hex
*
RDHEX	PSHB			Save B
	BSR	RDNIB		Get nibble
	BCS	rdherr		Error
	LSLA			Shift over
	LSLA			Into high
	LSLA			nibble
	LSLA			for later
	TAB			Save for later
	BSR	RDNIB		Get next nibble
	BCS	rdherr		Error
	ABA			Get final result
rdherr	PULB			Restore B
	RTS
*
* Read a word(A:B)
*
RDWORD	BSR	RDHEX		Get byte
	BCS	rdnerr		Error
	PSHA			Save for later
	BSR	RDHEX		Get byte
	BCS	rdherr		Error
	TAB			Copy to B
	PULA			Restore A
	RTS
*
* Read an address range... On exit X=start, OPRADR=End
*
RDADDR	STX	OPRADR		Save default size
	BSR	RDWORD		Get start address
	BCS	Goerr1		Error
	PSHB			Save LOW
	PSHA			Save HIGH
	LDAA	#','		Separator
	BSR	WRCHR		Output
	BSR	RDWORD		Get End address
	BCC	rda1		Its OK
	CMPA	#' '-'0'	Space?
	BNE	Goerr1		Report error
	PULA			Restore HIGH
	PSHA			Resave
	ADDB	OPRADR+1	Adjust by default size
	ADCA	OPRADR		Adjust high as well
	PSHA			Save it
	JSR	WRWORD		Output
	PULA			Restore
rda1	STAA	OPRADR		Save end address
	STAB	OPRADR+1
	TSX			Point to it
	LDX	0,X		Get value
	PULA			Restore HIGH
	PULB			Restore LOW
	RTS
Goerr1	JMP	ERROR		Report an error
*
* Change register
*
CHGREG	JSR	RDCHRE		Get register name
	LDX	#ACC		Assume ACC
	CMPA	#'A'		Accumulator?
	BEQ	chgr1		Yes
	INX			Advance to
	INX			Index
	CMPA	#'I'		Index?
	BEQ	chgr1		Yes
	INX			Advance to
	INX			Stack pointer
	CMPA	#'S'		Sp?
	BEQ	chgr1		Yes
	INX			Advance to
	INX			Program counter
	CMPA	#'P'		PC
	BNE	Goerr1
chgr1	JSR	WRSPACE		Space over
	JSR	RDWORD		Get value
	BCS	Goerr1		Error
	STAA	0,X		Write HIGH
	STAB	1,X		Write LOW
	RTS
*
* Set/Display brakpoints
*
BREAKPT	JSR	RDWORD		Get address
	BCC	break5		Set a breakpoint
break1	CMPA	#' '-'0'	Space?
	BNE	Goerr1		No, error
	LDX	#BRKTAB		Point to breakpoint table
break2	STX	TEMP		Save for later
	LDAB	1,X		Get LOW
	BNE	break3		Show it
	LDAA	0,X		Get HIGH
	BEQ	break4		Don't show
break3	JSR	WRSPACE		Output space
	LDAA	0,X		Read HIGH
	JSR	WRWORD		Output
break4	LDX	TEMP		Get address
	INX			Advance
	INX			Advance
	INX			Advance
	CPX	#BRKEND		At end?
	BNE	break2		Do them all
	RTS
* Set a breakpoint
break5	BSR	LOOKBRK		Does it exist already?
	BCC	Goerr1		Duplicate
	PSHA			Save ACCA
	PSHB			Save ACCB
	CLRA			Zero HIGH
	CLRB			Zero LOW
	BSR	LOOKBRK		Look it up
	BCS	Goerr1		Table is full
	PULB			Restore HIGH
	PULA			Restore LOW
	STAA	0,X		Save HIGH
	STAB	1,X		Save LOW
	RTS
*
* Kill a breakpoint
*
KILLB	JSR	RDWORD		Get breakpoint
	BCS	break1		Report error
	BSR	LOOKBRK		Does it exist?
	BCS	Goerr1		No, error
	CLR	0,X		Zero HIGH
	CLR	1,X		Zero LOW
	RTS
*
* Lookup a breakpoint in the breakpoint table
*
LOOKBRK	LDX	#BRKTAB		Point to table
lookb1	CMPA	0,X		Is this it?
	BNE	lookb2		No, try next
	CMPB	1,X		Is this it?
	BEQ	lookb3		Yes, return
lookb2	INX			Skip to next
	INX			Skip to next
	INX			Skip to next
	CPX	#BRKEND		Are we at end?
	BNE	lookb1		No, keep going
	SEC			Indicate fail
lookb3	RTS
*
* Unassemble instructions
*
UNASM	LDX	#$000F		Default to 16 bytes
	JSR	RDADDR		Get address
unas1	BSR	DISASM		Disassemble
	JSR	CPXEND		At end
	BLO	unas1		Do them all
	RTS
*
* Disassemble instruction at X
*
DISASM	JSR	WRLFCR		New line
DISNL	STX	TEMP		Save address
	JSR	WRADDT		Output adress
	JSR	WRSP2		Space over
	LDAA	0,X		Get instruction
	JSR	WRHEX		Output
	JSR	WRSP3		Space over
	LDAA	0,X		Get instruction
	CMPA	#$C8		Memory referance
	BHS	disa1		No, try next
* Lookup instruction in table
	LSRA			Drop 1 bit
	LSRA			Drop 1 bit
	LSRA
	ANDA	#%00011111	Get opcode
	LDX	#ITAB1		Point to table
	BSR	disins		Output instruction
	LDX	TEMP		Get address back
	LDAA	0,X		Get opcode back
	INX			Skip opcode
	TAB			Save for later
	ANDA	#%00000111	Save only addressing mode
* Immediate mode
	BNE	disop1		Not immediate
	LDAA	#'#'		Immediate
	JSR	WRCHR		Output
	BITB	#%00001000	Byte instruction
	BNE	disob		Yes, only one byte
	CMPB	#$98		Special?
	BHS	disob		Yes, only one byte
* Output a word
disow	LDAA	1,X		Get HIGH
	LDAB	0,X		Get LOW
	INX			Skip HIGH
	INX			Skip LOW
	JMP	WRWORD		Output
* Direct memory address
disop1	DECA			Direct?
	BEQ	disow		Output address
* Indirect through IDX
	DECA			Indirect?
	BEQ	disop2		Yes, output
* Indirect through Index with offset
	DECA			Indirect with offset?
	BNE	disop3		No, keep going
	BSR	disob		Output byte operand
	LDAA	#','		Separator
	JSR	WRCHR		Output
disop2	LDAA	#'I'		Indicate index
	JMP	WRCHR		Output
* Indirect through SP with offset
disop3	DECA			Indirect via SP?
	BNE	disop4		No, try next
	BSR	disob		Output operand
	STX	TEMP		Save for later
	LDX	#MISP		Point to message
	BRA	disop5		Output & exit
* Top of stack (remove)
disop4	STX	TEMP		Save for later
	LDX	#MTOS		Tops of stack (remove)
	DECA			Is this it?
	BEQ	disop5		Yes, handle it
	LDX	#MTOSR		Top of stack (remove)
	DECA			Is this it?
	BEQ	disop5		Correct
	LDX	#MTOSK		Top of stack (keep)
disop5	JSR	WRSTR		Output string
	LDX	TEMP		Restore address
	RTS
disob	LDAA	0,X		Get LOW
	INX			Skip LOW
	JMP	WRHEX		Output
*
* Display instruction(A) from table(X)
*
disins	TSTA			First instr?
	BEQ	disin2
disin1	TST	0,X		End of text?
	INX			Advance
	BPL	disin1		No, try next
	DECA			Is this it?
	BNE	disin1		No, keep going
disin2	LDAB	#7		space to output
* Write a string within a certain field
WRSTSP	JSR	WRSTR		Output string
disin3	JSR	WRSPACE		Output space
	DECB			Reduce count
	BNE	disin3		Do all spaces
disin4	RTS
*
* Non memory referance instructions
*
disa1	SUBA	#$C8		Convert to zero offset
	LDX	#ITAB2		Table to use
	BSR	disins		Output instriction
	LDX	TEMP		Get address back
	LDAA	0,X		Get opcode back
	INX			Skip opcode
* Handle extended address instructions
	CMPA	#$D8		CALL?
	BEQ	disow		Append address
	CMPA	#$D0		JMP?
	BLO	disa2		No, try next
	CMPA	#$D2		JNZ?
	BLS	disow		Append address
disa2	CMPA	#$DA		ALLOC?
	BEQ	disob
	CMPA	#$DB		FREE?
	BEQ	disob
	CMPA	#$EA		OUT?
	BEQ	disob
	CMPA	#$EB		IN?
	BEQ	disob
* Handle relative branch instruction
	CMPA	#$D3		SJMP?
	BLO	disin4		No, try next
	CMPA	#$D5		SJNZ?
	BHI	disin4		No, try next
	CLRA			Zero high
	LDAB	0,X		Get operand
	BPL	disa3		Not negative
	DECA			Sign extend
disa3	INX			Skip operand
	STX	TEMP		Save for later
	ADDB	TEMP+1		Compute address
	ADCA	TEMP		Compute address
	JMP	WRWORD		Output
*
* Display command help summary
*
HELP	LDX	#HELPTXT	Point to help text
HELP1	JSR	WRLFCR		New line
	LDAB	#20		Width to output
	BSR	WRSTSP		Output command
	JSR	WRSTR		Output descrition
	LDAA	0,X		At end?
	BNE	HELP1		No, keep going
	RTS
* Strings to output while disassembling operands
MISP	FCCH	',S'
MTOS	FCCH	'S+'
MTOSR	FCCH	'[S+]'
MTOSK	FCCH	'[S]'
* Table of memory referance instruction
ITAB1	FCCH	'LD'		00
	FCCH	'LDB'		08
	FCCH	'ADD'		10
	FCCH	'ADDB'		18
	FCCH	'SUB'		20
	FCCH	'SUBB'		28
	FCCH	'MUL'		30
	FCCH	'MULB'		38
	FCCH	'DIV'		40
	FCCH	'DIVB'		48
	FCCH	'AND'		50
	FCCH	'ANDB'		58
	FCCH	'OR'		60
	FCCH	'ORB'		68
	FCCH	'XOR'		70
	FCCH	'XORB'		78
	FCCH	'CMP'		80
	FCCH	'CMPB'		88
	FCCH	'LDI'		90
	FCCH	'LEAI'		98
	FCCH	'ST'		A0
	FCCH	'STB'		A8
	FCCH	'STI'		B0
	FCCH	'SHR'		B8
	FCCH	'SHL'		C0
* Table of standard instructions
ITAB2	FCCH	'LT'		C8
	FCCH	'LE'		C9
	FCCH	'GT'		CA
	FCCH	'GE'		CB
	FCCH	'ULT'		CC
	FCCH	'ULE'		CD
	FCCH	'UGT'		CE
	FCCH	'UGE'		CF
	FCCH	'JMP'		D0
	FCCH	'JZ'		D1
	FCCH	'JNZ'		D2
	FCCH	'SJMP'		D3
	FCCH	'SJZ'		D4
	FCCH	'SJNZ'		D5
	FCCH	'IJMP'		D6
	FCCH	'SWITCH'	D7
	FCCH	'CALL'		D8
	FCCH	'RET'		D9
	FCCH	'ALLOC'		DA
	FCCH	'FREE'		DB
	FCCH	'PUSHA'		DC
	FCCH	'PUSHI'		DD
	FCCH	'TAS'		DE
	FCCH	'TSA'		DF
	FCCH	'CLR'		E0
	FCCH	'COM'		E1
	FCCH	'NEG'		E2
	FCCH	'NOT'		E3
	FCCH	'INC'		E4
	FCCH	'DEC'		E5
	FCCH	'TAI'		E6
	FCCH	'TIA'		E7
	FCCH	'ADAI'		E8
	FCCH	'ALT'		E9
	FCCH	'OUT'		EA
	FCCH	'IN'		EB
	FCCH	'NATIVE'	EC
	FCCH	'?'		ED
	FCCH	'?'		EE
	FCCH	'?'		EF
	FCCH	'?'		F0
	FCCH	'?'		F1
	FCCH	'?'		F2
	FCCH	'?'		F3
	FCCH	'?'		F4
	FCCH	'?'		F5
	FCCH	'?'		F6
	FCCH	'?'		F7
	FCCH	'?'		F8
	FCCH	'?'		F9
	FCCH	'?'		FA
	FCCH	'?'		FB
	FCCH	'?'		FC
	FCCH	'?'		FD
	FCCH	'?'		FE
	FCCH	'?'		FF
* Table of register names and addresses
RDISTXT	FCCH	'A='
	FDB	ACC
	FCCH	' I='
	FDB	IDX
	FCCH	' S='
	FDB	SP
	FCCH	' P='
	FDB	PC
	FCB	0
* Misc strings.
HELLO	FCCH	"C-FLEA:"
TRAPMSG	FCCH	"*TRAP: "
BRKMSG	FCCH	"*BREAK: "
* Text to output in help response
HELPTXT	FCCH	'B [addr]'
	FCCH	'Set brkpt'
	FCCH	'C R value'
	FCCH	'Change register'
	FCCH	'D addr,[addr]'
	FCCH	'Dump memory'
	FCCH	'E addr'
	FCCH	'Edit memory'
	FCCH	'G [addr]'
	FCCH	'Go (execute)'
	FCCH	'K addr'
	FCCH	'Kill brkpt'
	FCCH	'R'
	FCCH	'Display regs'
	FCCH	'S'
	FCCH	'Single step'
	FCCH	'U addr,[addr]'
	FCCH	'Unassemble'
	FCCH	'X'
	FCCH	'Exit'
	FCCH	'Z'
	FCCH	'Reset CFLEA'
	FCB	0
*
* GO at full speed
*
GO	JSR	RDWORD		Get word
	BCC	go1		Set PC
	CMPA	#' '-'0'
	BEQ	go2		Allow it
	JMP	ERROR		Report error
go1	STAA	PC		Set PC LOW
	STAB	PC+1		Set PC HIGH
go2	JSR	WRLFCR		New line
	JSR	FETCH		Step one instruction
* Install breakpoints
	LDX	#BRKTAB		Get breakpoint table
go3	STX	TEMP		Save for later
	LDX	0,X		Get address
	BEQ	go4		Skip it
	LDAA	0,X		Get code from table
	LDAB	#$FF		Get TRAP
	STAB	0,X		Write it
	LDX	TEMP		Restore address
	STAA	2,X		Write to table
go4	LDX	TEMP		Restore address
	INX			Skip addressH
	INX			Skip addressL
	INX			Skip opcode
	CPX	#BRKEND		At end?
	BNE	go3		Keep going
* Execute at full speed
RUN	BSR	FETCH		Execute one inst
	BRA	RUN		And proceed
*
* Single step 1 instruction
*
STEP	LDX	PC		Get PC address
	JSR	DISNL		Disassemble
*
* Fetch and execute one CFLEA instruction
*
FETCH	LDX	PC		Get PC
	LDAA	0,X		Get opcode
	INX			Advance
	STX	PC		Resave PC
	CMPA	#$98		LD-LDI?
	BHS	Fetch10		No, try special memory reference
	BSR	GETADR		Get addressing mode
* Read data at address
	CLRA			Assume byte (high=0)
	LDAB	TEMP		Get opcode back
	BITB	#%00001000	Byte operation?
	BNE	Fetch2		No, its OK
	BITB	#%00000111	Immediate mode
	BNE	Fetch1		No, do not adjust
	INX			Advance PC
	STX	PC		Resave
Fetch1	LDX	OPRADR		Get address
	LDAA	1,X		Read HIGH value
Fetch2	STAA	TEMP		Save for later
	LDX	OPRADR		Get address
	LDAA	0,X		Read LOW value
	STAA	TEMP+1		Save for later
	LSRB			Drop addressing mode
	LSRB			Drop addressing mode
	LSRB			Drop addressing mode
	ANDB	#%00011110	Drop byte/word
	ADDB	#ITABLE1	Offset to table
	LDAA	#=ITABLE1	Get High address
	ADCA	#0		Adjust for overflow
	STAA	OPRADR		Save for copy
	STAB	OPRADR+1	To X
	LDX	OPRADR		Copy for indirect
	LDX	0,X		Get address at X
	LDAA	ACC		Get ACC HIGH
	LDAB	ACC+1		Get ACC LOW
	JMP	0,X
*
* Handle special memory referance instructions
*
Fetch10	CMPA	#$C8		A special memory instruction
	BHS	Fetch20		No, keep looking
	BSR	GETADR		Get addressing mode
	LDAB	TEMP		Get opcode back
	SUBB	#$98		Adjust to zero offset
	LSRB			Drop addressing mode
	LSRB			Drop addressing mode
	ANDB	#%00001110	Drop addressing mode
	ADDB	#ITABLE2	Offset to table
	LDAA	#=ITABLE2	Get high
	ADCA	#0		Adjust for carry
	STAA	TEMP		Save for copy
	STAB	TEMP+1		to X
	LDX	TEMP		Copy for indirect
	LDX	0,X		Get address at X
	JMP	0,X		Execute
*
* Other opcodes
*
Fetch20	SUBA	#$C8		Adjust to zero offset
	LSLA			x2 for word entries
	ADDA	#ITABLE3	Point to table
	LDAB	#=ITABLE3	Get HIGH
	ADCB	#0		Adjust for carry
	STAB	TEMP
	STAA	TEMP+1
	LDX	TEMP
	LDX	0,X		Get address
	JMP	0,X		Execute
*
* Determine addressing mode
*
GETADR	STAA	TEMP		Save for later
	ANDA	#%00000111	Save only addressing mode
* Immediate addressing mode
	BNE	geta1		Not immediate
	STX	OPRADR		Operand is at PC+
	BRA	geta31		Save
* Extended addressing
geta1	DECA			Extended?
	BNE	geta2		Not extended
	LDAA	0,X		Get LOW byte
	STAA	OPRADR+1	Save
	INX			Advance
	LDAA	0,X		Get HIGH byte
	STAA	OPRADR		Save
	INX			Advance
	STX	PC		Save new PC
	RTS
* Indirect through index register (no offset)
geta2	DECA			Indirect?
	BNE	geta3		Not indirect
	LDX	IDX		Get index register
	STX	OPRADR		Save
	RTS
* Indirect through index register (with offset)
geta3	DECA			Indirect/Offset?
	BNE	geta4		Not indirect
	LDAA	0,X		Get Offset
	ADDA	IDX+1		Add to index
	STAA	OPRADR+1	Save LOW
	LDAA	IDX		Get high
	ADCA	#0		Adjust for carry
	STAA	OPRADR		Save address
geta31:	INX			Advance
	STX	PC		Save new PC
	RTS
* Indirect through stack pointer (with offset)
geta4	DECA			Stack?
	BNE	geta5		No, try next
	LDAA	0,X		Get offset
	ADDA	SP+1		Add to stack
	STAA	OPRADR+1	Save LOW
	LDAA	SP		Get HIGH
	ADCA	#0		Adjust
	STAA	OPRADR		Resave
	BRA	geta31		Advance, save & exit
* Top of stack (remove)
geta5	DECA			Top of stack?
	BNE	geta6		No, try next
	LDX	SP		Get SP
	STX	OPRADR		Set operand address
geta51	INX			Advance
	INX			Advance
	STX	SP		Set new SP
	RTS
* Indirect through top of stack (remove)
geta6	DECA			Itop stack (leave)?
	BNE	geta7		No, try next
	BSR	geta7		Get address
	BRA	geta51		Advance, save & exit
* Indirect through top of stack (leave on)
geta7	LDX	SP		Get SP
	LDAA	0,X		Get LOW value
	STAA	OPRADR+1	Set operand address
	LDAA	1,X		Get HIGH value
	STAA	OPRADR		Set operand address
	RTS
*
* Standard memory referance instruction handlers
*
ITABLE1	FDB	_LD		00
	FDB	_ADD		10
	FDB	_SUB		20
	FDB	_MUL		30
	FDB	_DIV		40
	FDB	_AND		50
	FDB	_OR		60
	FDB	_XOR		70
	FDB	_CMP		80
	FDB	_LDI		90
	FDB	_LEAI		98
* Get flags from last compare operation
GETFLAG	LDAB	OPRADR		Get old flags
	ANDB	#%00001111	Mask old
	STAB	OPRADR		Resave
	TPA
	ANDA	#%11110000	Save flags
	ORAA	OPRADR		Include old
	TAP			Set new flags
	RTS
_NOT	LDAA	ACC		Get ACC
	ORAA	ACC+1		Zero?
	BEQ	RET1		Yes, return 1
	BRA	RET0		No, return 0
* Perform a compare operation... Save flags for later
_CMP	SUBB	TEMP+1		Subtract LOW operand
	STAB	ACC		Save LOW result
	SBCA	TEMP		Subtract HIGH operand
	TPA			A.3=S, A.2=Z, A.1=V, A.0=C
	TST	ACC		Test for really zero
	BEQ	cmp1		Yes, its OK
	ANDA	#%11111011	Clear 'Z' bit
cmp1	STAA	OPRADR		Save condition codes
	BITA	#%00000100	Zero?
	BEQ	RET0		Not same
RET1	LDAB	#1		Get 1
	BRA	CLAH		And save it
_LT	BSR	GETFLAG
	BLT	RET1
	BRA	RET0
_LE	BSR	GETFLAG
	BLE	RET1
	BRA	RET0
_GT	BSR	GETFLAG
	BGT	RET1
	BRA	RET0
_GE	BSR	GETFLAG
	BGE	RET1
	BRA	RET0
_ULT	BSR	GETFLAG
	BLO	RET1
	BRA	RET0
_ULE	BSR	GETFLAG
	BLS	RET1
	BRA	RET0
_UGT	BSR	GETFLAG
	BHI	RET1
	BRA	RET0
_UGE	BSR	GETFLAG
	BHS	RET1
RET0	CLRB			Zero LOW
CLAH	CLRA			Zero HIGH
stacc	STAB	ACC+1		Set ACC low
	STAA	ACC		Set ACC high
	RTS
*
_LD	LDX	TEMP		Get operand
	STX	ACC		Set ACC
	RTS
_ADD	ADDB	TEMP+1		Get LOW
	ADCA	TEMP		Get HIGH
	BRA	stacc		Save
_SUB	SUBB	TEMP+1		Sub LOW
	SUBA	TEMP		Sub HIGH
	BRA	stacc		Save
_AND	ANDB	TEMP+1		And LOW
	ANDA	TEMP		And HIGH
	BRA	stacc		Save
_OR	ORAB	TEMP+1		Or LOW
	ORAA	TEMP		Or HIGH
	BRA	stacc		Save
_XOR	EORB	TEMP+1		Xor LOW
	EORA	TEMP		Xor HIGH
	BRA	stacc		Save
_LDI	LDX	TEMP		Get TEMP
	STX	IDX		Set IDX
	RTS
* Multiply ACC and TEMP
_MUL	CLRA			Zero HIGH result
	CLRB			Zero LOW result
mul1	LSR	ACC		Shift HIGH n1
	ROR	ACC+1		Shift LOW n1
	BCC	mul2		No carry
	ADDB	TEMP+1		Add LOW n2
	ADCA	TEMP		Add HIGH n2
mul2	LDX	ACC		Test n1
	BEQ	stacc		End
	LSL	TEMP+1		Shift LOW n2
	ROL	TEMP		Shift HIGH n2
	LDX	TEMP		Test n2
	BNE	mul1		No, keep going
	BRA	stacc		Save ACC
* Divide ACC by TEMP ACC=Result
_DIV	LDX	#17		Bits to shift
	CLRA			Zero HIGH result
	CLRB			Zero LOW result
div1	CLC			Clear carry in
div2	ROL	ACC+1		Shift LOW n1
	ROL	ACC		Shift HIGH n1
	DEX			Reduce count
	BEQ	div4		Return
	ROLB			Shift in LOW
	ROLA			Shift in HIGH
	CMPA	TEMP		Less?
	BLO	div1		Yes, carry 0
	BHI	div3		No, exit
	CMPB	TEMP+1		Less?
	BLO	div1		Yes, carry zero
div3	SUBB	TEMP+1		Subtract n2
	SUBA	TEMP		Subtract n1
	SEC			Carry in 1
	BRA	div2		And proceed
div4	STAA	ALTVAL		Save HIGH remainder
	STAB	ALTVAL+1	Save LOW remainder
	RTS
*
* Special memory referance instruction handlers
*
ITABLE2	FDB	_LEAI		98
	FDB	_ST		A0
	FDB	_STB		A8
	FDB	_STI		B0
	FDB	_SHR		B8
	FDB	_SHL		C0
*
_LEAI	LDX	OPRADR
	STX	IDX		Set index register
	RTS
_ST	LDX	OPRADR
	LDAA	ACC		Get ACC HIGH
	STAA	1,X		Write ACC HIGH
_STB	LDX	OPRADR
	LDAA	ACC+1		Get ACC LOW
	STAA	0,X		Write ACC LOW
	RTS
_STI	LDX	OPRADR
	LDAB	IDX+1		Get LOW index
	LDAA	IDX		Get HIGH index
	STAB	0,X		Write LOW
	STAA	1,X		Write HIGH
	RTS
_SHR	LDX	OPRADR
	LDAA	0,X		Get count
	BEQ	shr2		0, do nothing
shr1	LSR	ACC		Shift HIGH
	ROR	ACC+1		Shift LOW
	DECA			Reduce count
	BNE	shr1		Do all shift
shr2	RTS
_SHL	LDX	OPRADR
	LDAA	0,X		Get count
	BEQ	shr2		Zero, do nothing
shl1	LSL	ACC+1		Shift LOW
	ROL	ACC		Shift HIGH
	DECA			Reduce count
	BNE	shl1		Do all shift
	RTS
*
* Individual instruction handlers
*
ITABLE3	FDB	_LT		C8
	FDB	_LE		C9
	FDB	_GT		CA
	FDB	_GE		CB
	FDB	_ULT		CC
	FDB	_ULE		CD
	FDB	_UGT		CE
	FDB	_UGE		CF
	FDB	_JMP		D0
	FDB	_JZ		D1
	FDB	_JNZ		D2
	FDB	_SJMP		D3
	FDB	_SJZ		D4
	FDB	_SJNZ		D5
	FDB	_IJMP		D6
	FDB	_SWITCH		D7
	FDB	_CALL		D8
	FDB	_RTS		D9
	FDB	_ALLOC		DA
	FDB	_FREE		DB
	FDB	_PUSHA		DC
	FDB	_PUSHI		DD
	FDB	_TAS		DE
	FDB	_TSA		DF
	FDB	_CLR		E0
	FDB	_COM		E1
	FDB	_NEG		E2
	FDB	_NOT		E3
	FDB	_INC		E4
	FDB	_DEC		E5
	FDB	_TAI		E6
	FDB	_TIA		E7
	FDB	_ADAI		E8
	FDB	_ALT		E9
	FDB	_OUT		EA
	FDB	_IN		EB
	FDB	_NATIVE		EC
	FDB	_TRAP		ED
	FDB	_TRAP		EE
	FDB	_TRAP		EF
	FDB	_TRAP		F0
	FDB	_TRAP		F1
	FDB	_TRAP		F2
	FDB	_TRAP		F3
	FDB	_TRAP		F4
	FDB	_TRAP		F5
	FDB	_TRAP		F6
	FDB	_TRAP		F7
	FDB	_TRAP		F8
	FDB	_TRAP		F9
	FDB	_TRAP		FA
	FDB	_TRAP		FB
	FDB	_TRAP		FC
	FDB	_TRAP		FD
	FDB	_TRAP		FE
	FDB	_TRAP		FF
*
* Push two bytes on the stack
*
PUSH2	LDX	SP		Get SP
	DEX			Backup
	STAA	0,X		Save HIGH
	DEX			Backup
	STAB	0,X		Save LOW
	STX	SP		Resave SP
	RTS
_PUSHA	LDAA	ACC		Get HIGH
	LDAB	ACC+1		Get LOW
	BSR	PUSH2		Stack it
	RTS
_PUSHI	LDAA	IDX		Get HIGH
	LDAB	IDX+1		Get LOW
	BSR	PUSH2		Stack it
	RTS
_CALL	LDX	PC		Get PC
	INX			Skip operand
	INX			Skip operand
	STX	TEMP		Resave
	LDAA	TEMP		Get high
	LDAB	TEMP+1		Get LOW
	BSR	PUSH2		Pust word on stack
_JMP	LDX	PC		Get PC
	LDAA	0,X		Get LOW
	LDAB	1,X		Get HIGH
	STAA	PC+1		Set LOW
	STAB	PC		Set HIGH
	RTS
_JZ	LDAA	ACC		Get HIGH
	ORAA	ACC+1		Is it zero?
	BEQ	_JMP		Yes, take it
skip2:	LDX	PC
	INX			Skip
	INX			Skip
	STX	PC		Resave
	RTS
_JNZ	LDAA	ACC		Get HIGH
	ORAA	ACC+1		Is it zero?
	BNE	_JMP		No, take it
	BRA	skip2
_SJMP	LDX	PC
	CLRB			Zero high
	LDAA	0,X		Get byte value
	BPL	sjmp1		It's postive
	DECB			Sign extend
sjmp1	INX			Advance
	STX	PC		Set new address
	ADDA	PC+1		Add offset
	STAA	PC+1		Save new
	ADCB	PC		Do high
	STAB	PC		And resave
	RTS
_SJZ	LDAA	ACC		Get HIGH
	ORAA	ACC+1		Is it zero?
	BEQ	_SJMP		Yes, take it
skip1:	LDX	PC
	INX			Skip to next
	STX	PC		Resave
	RTS
_SJNZ	LDAA	ACC		Get HIGH
	ORAA	ACC+1		Is it zero?
	BNE	_SJMP		Yes, take it
	BRA	skip1
_IJMP	LDX	ACC		Get accumulator
	STX	PC		Save PC
	RTS
_SWITCH	LDX	IDX		Get index address
swit1	LDAA	0,X		Get address
	ORAA	1,X		At end?
	BEQ	swit3		Assume default
	LDAA	2,X		Get LOW value
	CMPA	ACC+1		Match?
	BNE	swit2		No, go to next
	LDAA	3,X		Get HIGH
	CMPA	ACC		Match?
	BEQ	swit4		No, go to next
swit2	INX			Skip address low
	INX			Skip address high
	INX			Skip data low
	INX			Skip data high
	BRA	swit1		And proceed
swit3	INX			Skip 0 (LOW)
	INX			Skip 0 (HIGH)
swit4	LDAA	0,X		Get LOW address
	STAA	PC+1		Set new HIGH
	LDAA	1,X		Get HIGH address
	STAA	PC		Set new LOW
	RTS
_RTS	LDX	SP		Get SP
	LDAA	0,X		Get LOW
	INX			Advance
	STAA	PC+1		Set PC
	LDAA	0,X		Get HIGH
	INX			Advance
	STAA	PC		Set PC
	STX	SP		Resave stack
	RTS
_ALLOC	LDX	PC
	LDAA	SP+1		Get LOW SP
	SUBA	0,X		Subtract value
	STAA	SP+1		Resave
	LDAA	SP		Get HIGH
	SBCA	#0		Adjust
	STAA	SP		Resave
	BRA	skip1		Advance PC
_FREE	LDX	PC
	LDAA	SP+1		Get LOW SP
	ADDA	0,X		Add value
	STAA	SP+1		Resave
	LDAA	SP		Get LOW
	ADCA	#0		Adjust
	STAA	SP		Resave
	BRA	skip1		Advance PC
_TAS	LDX	ACC		Get ACC
	STX	SP		Set SP
	RTS
_TSA	LDX	SP		Get SP
	STX	ACC		Save ACC
	RTS
_CLR	CLR	ACC		Zero HIGH
	CLR	ACC+1		Zero LOW
	RTS
_COM	COM	ACC		Complement HIGH
	COM	ACC+1		Complement LOW
	RTS
_NEG	COM	ACC		Complement HIGH
	COM	ACC+1		Complement LOW
_INC	LDX	ACC		Get ACC
	INX			Advance
	STX	ACC		Resave
	RTS
_DEC	LDX	ACC		Get ACC
	DEX			Backup
	STX	ACC		Resave
	RTS
_TAI	LDX	ACC		Get ACC
	STX	IDX		Save Index
	RTS
_TIA	LDX	IDX		Get Index
	STX	ACC		Save ACC
	RTS
_ADAI	LDAA	IDX+1		Get index LOW
	ADDA	ACC+1		Add ACC
	STAA	IDX+1		Resave
	LDAA	IDX		Get HIGH
	ADCA	ACC		Add ACC
	STAA	IDX		Resave
	RTS
_NATIVE	PULA			Get HIGH return
	STAA	OPRADR		Save it
	PULA			Get LOW return
	STAA	OPRADR+1	Save it
	LDAA	PC+1		Get PC LOW
	PSHA			Save it
	LDAA	PC		Get PC HIGH
	PSHA			Save it
	LDX	IDX		Get index
	LDAA	ACC		Get ACCH
	LDAB	ACC+1		GET ACCL
	RTS
_ALT	LDX	ALTVAL		Get alternate
	STX	ACC		Set ACC
	RTS
*
* Write to an output port
*
_OUT	LDX	PC		Get PC address
	LDAB	0,X		Get PORT ID
	INX			Advance
	STX	PC		Resave
	LDAA	ACC+1		Get data
* Port 0 - Write character to serial port
	TSTB			Port 0?
	BNE	inrts		No, ignore
	JMP	WRCHR		Output
*
* Read from an input port
*
_IN	LDX	PC		Get PC address
	LDAA	0,X		Get PORT ID
	INX			Advance
	STX	PC		Resave
* Port 0 - Read character from serial port
	TSTA			Port 0?
	BNE	inrts		No, try next
	JSR	RDCHR		Get a character
	STAA	ACC+1		Save value
	CLR	ACC		Zero extend
inrts	RTS
