;
; MICRO-C/PC callable C-FLEA Virtual Machine interpreter
;
; Assemble with MASM (or compatible): masm/ml VMPC86;
;
; Copyright 1994-2005 Dave Dunfield
; All rights reserved.
;
DGRP	GROUP	DSEG,BSEG
CSEG	SEGMENT	BYTE PUBLIC 'CODE'
	ASSUME	CS:CSEG, DS:DGRP, SS:DGRP
	PUBLIC	_cf_run
	EXTRN	_cf_output:NEAR, _cf_input:NEAR
;
; Addressing mode handler table
;
ADDRT	LABEL	WORD
	DW	?IMM		; Immediate mode
	DW	?EXT		; Extended mode
	DW	?IDX		; Through index (No offset)
	DW	?IDXO		; Through index (with offset)
	DW	?STKO		; Through index (with offset)
	DW	?TOS		; Top of stack (remove)
	DW	?ITOS		; Indirect through TOS (remove)
	DW	?ITOSK		; Indirect through TOS (keep)
;
; Memory regerencing instruction table
;
MEMOT	LABEL	WORD
	DW	?LD
	DW	?ADD
	DW	?SUB
	DW	?MUL
	DW	?DIV
	DW	?AND
	DW	?OR
	DW	?XOR
	DW	?CMP
	DW	?LDI
;
; Special memory referencing instructions
;
SMEMT	LABEL	WORD
	DW	?LEAI
	DW	?ST
	DW	?STB
	DW	?STI
	DW	?SHR
	DW	?SHL
;
; Other OPCODES
;
OPCT	LABEL	WORD
	DW	?LT		; C8
	DW	?LE		; C9
	DW	?GT		; CA
	DW	?GE		; CB
	DW	?ULT		; CC
	DW	?ULE		; CD
	DW	?UGT		; CE
	DW	?UGE		; CF
	DW	?JMP		; D0
	DW	?JZ		; D1
	DW	?JNZ		; D2
	DW	?SJMP		; D3
	DW	?SJZ		; D4
	DW	?SJNZ		; D5
	DW	?IJMP		; D6
	DW	?SWITCH		; D7
	DW	?CALL		; D8
	DW	?RTS		; D9
	DW	?ALLOC		; DA
	DW	?FREE		; DB
	DW	?PUSHA		; DC
	DW	?PUSHI		; DD
	DW	?TAS		; DE
	DW	?TSA		; DF
	DW	?CLR		; E0
	DW	?COM		; E1
	DW	?NEG		; E2
	DW	?NOT		; E3
	DW	?INC		; E4
	DW	?DEC		; E5
	DW	?TAI		; E6
	DW	?TIA		; E7
	DW	?ADAI		; E8
	DW	?ALT		; E9
	DW	?OUT		; EA
	DW	?IN		; EB
	DW	ILLEGAL		; EC
	DW	ILLEGAL		; ED
	DW	ILLEGAL		; EE
	DW	ILLEGAL		; EF
	DW	ILLEGAL		; F0
	DW	ILLEGAL		; F1
	DW	ILLEGAL		; F2
	DW	ILLEGAL		; F3
	DW	ILLEGAL		; F4
	DW	ILLEGAL		; F5
	DW	ILLEGAL		; F6
	DW	ILLEGAL		; F7
	DW	ILLEGAL		; F8
	DW	ILLEGAL		; F9
	DW	ILLEGAL		; FA
	DW	ILLEGAL		; FB
	DW	ILLEGAL		; FC
	DW	ILLEGAL		; FD
	DW	ILLEGAL		; FE
	DW	ILLEGAL		; FF
;
; Multi step several instructions
;
_cf_run:CALL	LOAD		; Load registers
cstep:	CALL	SINSTR		; Step the instruction
	JMP	SHORT cstep	; And keep going
;
; Illegal opcode detected, report
;
ILLEGAL:POP	AX		; Skip return address
	DEC	SI		; Backup to it
	MOV	AL,ES:[SI]	; Get illegal instruction
	XOR	AH,AH		; Zero high
	JMP	SAVE		; Save registers & exit
;
; Simulate a CFLEA instruction
; DX=ACC, SI=PC, ES=Code segment
;
; Note first byte of this routine is patched to 89B6 (MOV) or
; EBnn (JMP) by the execution routines and the INT-9 handler.
;
SINSTR:	MOV	BL,ES:[SI]	; Get opcode
	XOR	BH,BH		; Zero high
	INC	SI		; Advance PC
	CMP	BL,98h		; "normal" memory address?
	JAE	?step3		; No, try next
;
; Normal memory reference instruction
;
	CALL	GETADR		; Get address
	TEST	BL,0Fh		; Word immediate?
	JNZ	?step1		; Nope, its OK
	INC	SI		; Skip extra byte
; Read memory location into CX
?step1:	XOR	CH,CH		; Assume zero
	TEST	BL,08h		; Byte access
	JNZ	?step2		; Yes, its OK
	MOV	CH,ES:1[DI]	; Read HIGH
?step2:	MOV	CL,ES:[DI]	; Read a byte
	SHR	BL,1		; Shift over
	SHR	BL,1		; Shift over
	SHR	BL,1		; Shift over
	AND	BL,1Eh		; Save opcode
	JMP	MEMOT[BX]	; Execute handler
;
; Special memory-reference
;
?step3:	CMP	BL,0C8h		; Special addressing mode instructino
	JAE	?step4		; No, try next
	CALL	GETADR		; Get address of instruction
	SUB	BL,098h		; Adjust to zero offset
	SHR	BL,1		; Clear addressing mode
	SHR	BL,1		; Clear addrssing mode
	AND	BL,0Eh		; Save jump table
	JMP	SMEMT[BX]	; Handle special instructions
;
; Other instructions
;
?step4:	SUB	BL,0C8h		; Convert to zero offset
	SHL	BL,1		; X2 for two byte entries
	JMP	OPCT[BX]	; Execute handler
;
; Memory referencing instructions
;
?LD:	MOV	DX,CX		; Load accumulator
	RET
?ADD:	ADD	DX,CX		; Add to accumulator
	RET
?SUB:	SUB	DX,CX		; Subtract from acc
	RET
?MUL:	MOV	AX,DX		; Get ACC
	MUL	CX		; Multiply
	MOV	DGRP:_TEMP,DX	; Save high result
	MOV	DX,AX		; Get low result
	RET
?DIV:	MOV	AX,DX		; Get low
	XOR	DX,DX		; Zero high
	AND	CX,CX		; Zero is invalid
	JZ	div1		; Return zeros
	DIV	CX		; Divide
	MOV	DGRP:_TEMP,DX	; Save remainder
	MOV	DX,AX		; Get Quotient
	RET
div1:	MOV	DGRP:_TEMP,DX	; Zero ALT
	RET
?AND:	AND	DX,CX		; Perform and
	RET
?OR:	OR	DX,CX		; Perform OR
	RET
?XOR:	XOR	DX,CX		; Perform XOR
	RET
?CMP:	CMP	DX,CX		; Perform compare
	PUSHF			; Save flags
	POP	WORD PTR DGRP:FLAGS; Save 8086 flags
	JZ	?ret1		; Values are the same
?ret0:	XOR	DX,DX		; Zero DX
	RET
?NOT:	AND	DX,DX		; Test ACC != 0
	JNZ	?ret0		; Yes, return 0
?ret1:	MOV	DX,1		; Return 1
	RET
?LT:	PUSH	WORD PTR DGRP:FLAGS; Save 8086 flags
	POPF			; Get flags
	JL	?ret1		; TRUE
	XOR	DX,DX		; Return zero
	RET
?LE:	PUSH	WORD PTR DGRP:FLAGS; Save 8086 flags
	POPF			; Get flags
	JLE	?ret1		; TRUE
	XOR	DX,DX		; Return zero
	RET
?GT:	PUSH	WORD PTR DGRP:FLAGS; Save 8086 flags
	POPF			; Get flags
	JG	?ret1		; TRUE
	XOR	DX,DX		; Return zero
	RET
?GE:	PUSH	WORD PTR DGRP:FLAGS; Save 8086 flags
	POPF			; Get flags
	JGE	?ret1		; TRUE
	XOR	DX,DX		; Return zero
	RET
?ULT:	PUSH	WORD PTR DGRP:FLAGS; Save 8086 flags
	POPF			; Get flags
	JB	?ret1		; TRUE
	XOR	DX,DX		; Return zero
	RET
?ULE:	PUSH	WORD PTR DGRP:FLAGS; Save 8086 flags
	POPF			; Get flags
	JBE	?ret1		; TRUE
	XOR	DX,DX		; Return zero
	RET
?UGT:	PUSH	WORD PTR DGRP:FLAGS; Save 8086 flags
	POPF			; Get flags
	JA	?ret1		; TRUE
	XOR	DX,DX		; Return zero
	RET
?UGE:	PUSH	WORD PTR DGRP:FLAGS; Save 8086 flags
	POPF			; Get flags
	JAE	?ret1		; TRUE
	XOR	DX,DX		; Return zero
	RET
?LDI:	MOV	DGRP:_INDEX,CX	; Set index register
	RET
;
; Special memory referencing instructions
;
?LEAI:	MOV	DGRP:_INDEX,DI	; Set address
	RET
?ST:	MOV	ES:1[DI],DH	; Save HIGH value
?STB:	MOV	ES:[DI],DL	; Save LOW value
	RET
?STI:	MOV	AX,DGRP:_INDEX	; Get index
	MOV	ES:[DI],AL	; Write LOW
	MOV	ES:1[DI],AH	; Write HIGH
	RET
?SHR:	MOV	CL,ES:[DI]	; Get count
	SHR	DX,CL		; Shift it
	RET
?SHL:	MOV	CL,ES:[DI]	; Get count
	SHL	DX,CL		; Shift it
	RET
;
; Other instructions
;
?JNZ:	AND	DX,DX		; Test ACC != 0
	JZ	?jmp1		; No, don't do it
?JMP:	MOV	AL,ES:[SI]	; Get LOW
	MOV	AH,ES:1[SI]	; Get HIGH
	MOV	SI,AX		; Set new PC
	RET
?JZ:	AND	DX,DX		; Test ACC == 0
	JZ	?JMP		; Yes, jump
?jmp1:	INC	SI		; Skip low
	INC	SI		; Skip high
	RET
?SJNZ:	AND	DX,DX		; Test ACC != 0
	JZ	?sjmp1		; No, don't do it
?SJMP:	MOV	AL,ES:[SI]	; Get offset
	INC	SI		; Advance
	CBW			; Sign extend
	ADD	SI,AX		; Adjust to new
	RET
?SJZ:	AND	DX,DX		; Test ACC == 0
	JZ	?SJMP		; Yes, jump
?sjmp1:	INC	SI		; Skip offset
	RET
?IJMP:	MOV	SI,DX		; Set new PC
	RET
?SWITCH:MOV	DI,DGRP:_INDEX	; Get index register
?swit1:	MOV	BL,ES:[DI]	; Get LOW address
	INC	DI		; Skip LOW
	MOV	BH,ES:[DI]	; Get HIGH address
	INC	DI		; Skip HIGH
	MOV	AL,ES:[DI]	; Get LOW value
	INC	DI		; Skip LOW
	MOV	AH,ES:[DI]	; Get HIGH value
	INC	DI		; Skip HIGH
	AND	BX,BX		; Zero?
	JZ	?swit2		; Yes, do default
	CMP	AX,DX		; Does it match
	JNZ	?swit1		; Keep looking
	MOV	AX,BX		; Get address
?swit2:	MOV	SI,AX		; Jump to default
	RET
?CALL:	MOV	CL,ES:[SI]	; Get LOW address
	INC	SI		; Advance
	MOV	CH,ES:[SI]	; Get HIGH address
	INC	SI		; Advance
	MOV	AX,SI		; Get PC
	CALL	PUSH16		; Save it
	MOV	SI,CX		; Set new value
	RET
?RTS:	MOV	DI,DGRP:_SP	; Get SP
	MOV	AL,ES:[DI]	; Get LOW value
	INC	DI		; Remove LOW
	MOV	AH,ES:[DI]	; Get HIGH value
	INC	DI		; Remove HIGH
	MOV	DGRP:_SP,DI	; Resave
	MOV	SI,AX		; Set new PC
	RET
?ALLOC:	MOV	DI,DGRP:_SP	; Get SP
	MOV	AL,ES:[SI]	; Get value
	INC	SI		; Skip
	XOR	AH,AH		; Zero high
	SUB	DI,AX		; Allocate space
	MOV	DGRP:_SP,DI	; Resave
	RET
?FREE:	MOV	AL,ES:[SI]	; Get value
	INC	SI		; Skip
	XOR	AH,AH		; Zero HIGH
	ADD	AX,DGRP:_SP	; Add to SP
	MOV	DGRP:_SP,AX	; Resave
	RET
?PUSHA:	MOV	AX,DX		; Get ACC
PUSH16:	MOV	DI,DGRP:_SP	; Get SP
	DEC	DI		; Backup
	MOV	ES:[DI],AH	; Save HIGH value
	DEC	DI		; Backup
	MOV	ES:[DI],AL	; Save LOW value
	MOV	DGRP:_SP,DI	; Resave
	RET
?PUSHI:	MOV	AX,DGRP:_INDEX	; Get index
	JMP	SHORT PUSH16	; And save
?TAS:	MOV	DGRP:_SP,DX	; Save ACC
	RET
?TSA:	MOV	DX,DGRP:_SP	; Get SP
	RET
?CLR:	XOR	DX,DX		; Zero ACC
	RET
?COM:	NOT	DX		; Complement
	RET
?NEG:	NEG	DX		; Negate
	RET
?INC:	INC	DX		; Increment
	RET
?DEC:	DEC	DX		; Decrement
	RET
?TAI:	MOV	DGRP:_INDEX,DX	; Save ACC
	RET
?TIA:	MOV	DX,DGRP:_INDEX	; Get index
	RET
?ADAI:	MOV	AX,DGRP:_INDEX	; Get index
	ADD	AX,DX		; Adjust by ACC
	MOV	DGRP:_INDEX,AX	; Resave
	RET
?ALT:	MOV	DX,DGRP:_TEMP	; Get temp vale
	RET
;
; Handle OUTPUT command
;
?OUT:	MOV	AL,ES:[SI]	; Get port number
	INC	SI		; Skip operand
	CALL	SAVE		; Save registers
	XOR	AH,AH		; Zero high
	PUSH	AX		; Pass port address
	CALL	_cf_output	; Perform output
	POP	AX		; Clean stack
	JMP	LOAD		; Restore registers
;
; Read from I/O port
;
?IN:	MOV	AL,ES:[SI]	; Get port number
	INC	SI		; Skip to next
	CALL	SAVE		; Save registers
	XOR	AH,AH		; Zero high
	PUSH	AX		; Pass parameter
	CALL	_cf_input	; Perform input
	POP	AX		; Clean stack
	JMP	LOAD		; Restore registers
;
; Get address from addressing mode (output in DI)
;
GETADR:	MOV	DI,BX		; Copy
	SHL	DI,1		; Shift it
	AND	DI,000Eh	; Save only mode bits
	JMP	ADDRT[DI]	; Execute handler
?IMM:	MOV	DI,SI		; Point to PC
	INC	SI		; Advance PC
	RET
?EXT:	MOV	AL,ES:[SI]	; Get LOW address
	INC	SI		; Advance PC
	MOV	AH,ES:[SI]	; Get HIGH address
	INC	SI		; Advance PC
	MOV	DI,AX		; Set new address
	RET
?IDX:	MOV	DI,DGRP:_INDEX	; Get index register
	RET
?IDXO:	MOV	AL,ES:[SI]	; Get offset
	XOR	AH,AH		; Zero HIGH
	ADD	AX,DGRP:_INDEX	; Offset to index
	MOV	DI,AX		; Save address
	INC	SI		; Advance PC
	RET
?STKO:	MOV	AL,ES:[SI]	; Get offset
	XOR	AH,AH		; Zero high
	ADD	AX,DGRP:_SP	; Offset to stack
	MOV	DI,AX		; Save address
	INC	SI		; Advance PC
	RET
?TOS:	MOV	AX,DGRP:_SP	; Get stack pointer
	MOV	DI,AX		; Point to it
	INC	AX		; Advance
	INC	AX		; Advance
	MOV	DGRP:_SP,AX	; Resave
	RET
?ITOS:	MOV	DI,DGRP:_SP	; Get stack pointer
	MOV	AL,ES:[DI]	; Get LOW address
	INC	DI		; Advance
	MOV	AH,ES:[DI]	; Get HIGH address
	INC	DI		; Advance
	MOV	DGRP:_SP,DI	; Resave SP
	MOV	DI,AX		; Get address
	RET
?ITOSK:	MOV	DI,DGRP:_SP	; Get stack pointer
	MOV	AL,ES:[DI]	; Get LOW address
	MOV	AH,ES:1[DI]	; Get HIGH address
	MOV	DI,AX		; Set address
	RET
;
; Load pseudo-registers
;
LOAD:	MOV	DX,DGRP:_ACC	; Get ACC
	MOV	SI,DGRP:_PC	; Get PC
	MOV	ES,DGRP:_CFSEG	; Get memory segment
	RET
;
; Save pseudo-registers
;
SAVE:	MOV	DGRP:_ACC,DX	; Save ACC
	MOV	DGRP:_PC,SI	; Save PC
	RET
CSEG	ENDS
;
; Initialized data
;
DSEG	SEGMENT	BYTE PUBLIC 'IDATA'
	PUBLIC	_ACC, _INDEX, _SP, _PC
_ACC	DW	0		; A Accumulator
_INDEX	DW	0		; Index register
_SP	DW	256		; Stack Pointer
_PC	DW	0		; Program Counter
_TEMP	DW	0		; Temporary output
FLAGS	DW	0200h		; 8086 Flags
DSEG	ENDS
;
; Uninitialized data
;
BSEG	SEGMENT	BYTE PUBLIC 'UDATA'
	PUBLIC	_CFSEG
_CFSEG	DW	1 DUP(?)	; Active code (ROM) segment
BSEG	ENDS
	END
